/**
* @author Dirk Nordmann
* @date   14.01.2002
*/
package xml2rtf.rtf;

import util.*;
import xml2rtf.xml.*;
import xml2rtf.templates.*;
import java.io.*;
import java.util.Stack;
import java.util.Vector;
import java.util.StringTokenizer;

public class RtfGenerator
{

  private RtfWriter _output;
  private XmlReader _input;
  private String _text;
  private RtfTable _currentTable;
  private RtfGeneratorContext _currentContext;
  private boolean currentLineClosed;
  private boolean previousLineIsEmpty;
  private boolean _tableActive;
  private boolean tableWasActive;
  private boolean _smallcapsActive;
  private boolean _hyperlinkActive;
  private Stack tableStack;
  private Stack _contextStack;
  private Stack _savedContextStack;
  private boolean nextWordOnBol;

  /**
  * Constructor
  */
  public RtfGenerator()
  {
  }
	
  /**
  * Returns the input xml-reader for this generator.
  * @return XmlReader
  */
  public XmlReader getInput()
  {
    return _input;
  }

  /**
  * Returns the output rtf-writer for this generator.
  * @return XmlReader
  */
  public RtfWriter getOutput()
  {
    return _output;
  }

	
	
  /**
  * generate
  */
  public void generate(RtfWriter rtfwriter, XmlReader xmlreader)
  {
    _output = rtfwriter;
  	_input = xmlreader;
    _text = "";
  	_currentTable = null;
    currentLineClosed = true;
  	previousLineIsEmpty = true;
    _tableActive = false;
  	tableWasActive = false;
    tableStack = new Stack();
  	_currentContext = new RtfGeneratorContext();
    _contextStack = new Stack();
  	_savedContextStack = new Stack();
    _smallcapsActive = false;
  	nextWordOnBol = true;
    _hyperlinkActive = false;
  	resetToTopOfPage();
    //* generate rtf *
    Parse(xmlreader);
  }

  /**
  * Parse
  */
  private void Parse(XmlTokenizer xmltokenizer)
  {
    String s;
    while ((s = xmltokenizer.getNextToken())!=null)
    {
    	if (s != null)
    	{
			//***** Tag *****
    		if (s.equals("<"))
    		{
    			//----- Tag-Content -----
    			String tag = s;
				String _s;
			    boolean inQuotes = false;
    			while ((_s = xmltokenizer.getNextToken())!=null)
    			{
        			boolean inComment = tag.startsWith("<!--");
    				//* Tag content *
    				StringTokenizer st = new StringTokenizer(_s, "\"'", true);
					while (st.hasMoreTokens()) 
        			{
						String __s = st.nextToken();
						tag += __s;
						// String in quotes 
						if(!inComment && (__s.equals("\"") || __s.equals("'")))
							inQuotes = !inQuotes;
        			}
        			//* Close tag *
        			boolean close = !inQuotes && tag.endsWith(">");
        			//* Close comment tag *
        			if (inComment)
        				close = tag.endsWith("-->");
        			//* Break on close tag *
    				if (close)
    					break;
    			}  
    			//----- Process Tag -----
    			if (tag.startsWith("<") && tag.endsWith(">"))
                	processTag(new Tag(tag));
				//----- Error: not well-formed -----
    			else 
				{
    				System.out.println("  tag='" + tag + "'");
    				System.out.println("*****[RtfGenerator.Parse]: FEHLER - Tag ungltig!*****");
    				System.exit(1);
    			}
    		}
    		//***** Text *****
    		else {
				//----- Process text -----
  	  			if (s.length() > 0)
    			{
	    	    	Globals.writeDebug("[RtfGenerator.Parse]: TOKEN_TEXT s='" + s + "'");
  	    	    if (_tableActive)
                _currentTable.addTextContentsToCell(HtmlObject.translateSpecialChars(s, true), getJavaFontSize(_currentContext.font_size), _currentContext.isBold, _currentContext.isItalic, _currentContext.isFixedFont);
					_text += s;
				}
			}
		}
	}
    		
    //*****  Dump  *****
    dumpText();
  }

  /**
  * getFontSize
  * @param  i
  * @return int
  */        
  private int getFontSize(int i)
  {
    if (i <= 2)
      return 15;
  	if (i == 3)
    	return 18;
  	if (i == 4)
    	return 21;
  	if (i == 5)
    	return 25;
  	return i != 6 ? 32 : 28;
  }

  /**
  * getJavaFontSize
  * @param  i
  * @return int
  */        
  private int getJavaFontSize(int i)
  {
    if (i<=2)
      return 10;
  	if (i==3)
    	return 12;
  	if (i==4)
    	return 14;
  	if (i==5)
    	return 17;
  	return i != 6 ? 21 : 19;
  }

private void indent()
{
	_currentContext.list_offset += ConversionTemplate.LST_OFFSET;
	outputRtf(RtfWriter.PARD + "\\li" + _currentContext.list_offset + " ");
}

private void desindent()
{
	_currentContext.list_offset -= ConversionTemplate.LST_OFFSET;
	if(_currentContext.list_offset < 0)
		_currentContext.list_offset = 0;
	outputRtf(RtfWriter.PARD + "\\li" + _currentContext.list_offset + " ");
}

private void processTag(Tag xmltag)
{
	Globals.writeDebug("[RtfGenerator.processTag]: " + xmltag._tag);
        
	//*****  Comments: <!-- ... --> *****
	if (xmltag.isComment())
	{
    Globals.writeDebug("[RtfGenerator.processTag]: Comment detected.");
    return;
	}
		
	RtfGeneratorContext Context = null;
	HtmlTagInfo taginfo = HtmlTag.findTag(xmltag.getTagName());
        
	if (taginfo==null)
		return;
            
	if(!xmltag.isStartTag() && !taginfo.needsCloseTag())
	{
		Globals.writeDebug("[RtfGenerator.processTag]: WARNING -- Unexpected " + taginfo.Text() + " close-tag ignored.");
		return;
	}
        
	for (Context = _contextStack.empty() ? null : (RtfGeneratorContext)_contextStack.peek(); 
		!xmltag.isStartTag() 
		&& Context != null 
		&& !taginfo.Text().equals(Context.tag_str); 
		Context = _contextStack.empty() ? null : (RtfGeneratorContext)_contextStack.peek())
	{
		if(!inStack(taginfo.Text()))
			return;
		processTag(new Tag("</" + Context.tag_str + ">"));
	}

	//*****  Dump  *****
	dumpText();
        
        if (taginfo.Ident() == HtmlTag.TABLE)
            if (xmltag.isStartTag())
            {
                if(_savedContextStack == null)
                {
                    _savedContextStack.push(_contextStack);
                    _contextStack = new Stack();
                }
            } else
            if(!_savedContextStack.empty())
                _contextStack = (Stack)_savedContextStack.pop();
                
        if (xmltag.isStartTag())
        {
            RtfGeneratorContext Context1;
            if(!_contextStack.empty())
                Context1 = (RtfGeneratorContext)_contextStack.peek();
            _currentContext.current_tag = taginfo.Ident();
            _currentContext.tag_str = taginfo.Text();
            Context1 = new RtfGeneratorContext(_currentContext);
            _contextStack.push(Context1);
        }
        
        switch (taginfo.Ident())
        {
        
        //------------------------------------
        //-----      Unhandled Tags      -----
      	//------------------------------------
        case HtmlTag.APPLET: 
        case HtmlTag.BGSOUND: 
       	case HtmlTag.BLINK:
        case HtmlTag.BLOCKQUOTE: 
        case HtmlTag.BODY:
       	case HtmlTag.MARQUEE:
       	case HtmlTag.PRE:
        case HtmlTag.TEXTAREA: 
            System.out.println("[RtfGenerator.processTag]: WARNUNG -- " + taginfo.Text() + "-Tag nicht verarbeitet.");
            break;

    		//--------------------------
     		//-----      Font      -----
      	//--------------------------
   			
        //***** FONT size=??? color=??? face=??? *****
        case HtmlTag.FONT: 
            if(xmltag.isStartTag())
            {
           		String rtf = "";
            		
           		//* FONT : size *
                String font_size = xmltag.getTagAttr("size");
                if(font_size != null)
                    try
                    {
                        int l;
                        if(font_size.charAt(0) == '+')
                            l = _currentContext.font_size + Integer.parseInt(font_size.substring(1));
                        else if(font_size.charAt(0) == '-')
                            l = _currentContext.font_size - Integer.parseInt(font_size.substring(1));
                        else
                            l = Integer.parseInt(font_size);
                        if(l < 1)
                            l = 1;
                        else if(l > 7)
                            l = 7;
                        _currentContext.font_size = l;
                    }
                    catch(NumberFormatException _ex)
                    {
                        System.out.println("[RtfGenerator.processTag]: WARNUNG -- ungltige Gre (font_size=" + font_size + ") fr FONT-Tag.");
                    }
                rtf += "{\\fs" + getFontSize(_currentContext.font_size) + " ";
                
                //* FONT : color *
                String font_color = xmltag.getTagAttr("color");
                if(font_color != null)
                    rtf += _output.getColor(font_color) + " ";
                
           		//* FONT : face *
                String font_face = xmltag.getTagAttr("face");
                if(font_face != null)
                {
                    font_face = font_face.toUpperCase();
                    if(font_face.indexOf("BOLD") >= 0)
                        rtf += "\\b ";
                    if(font_face.indexOf("ITALIC") >= 0)
                        rtf += "\\i ";
                }
               outputRtf(rtf);
               
            } else
               outputRtf("}");
               
            break;

        //***** Italics (kursiv) *****
        case HtmlTag.ADDRESS: 
        case HtmlTag.CITE: 
        case HtmlTag.DFN:
        case HtmlTag.EM:
        case HtmlTag.ITALIC:
        case HtmlTag.VAR:
            outputRtf(xmltag.isStartTag() ? "{\\i " : "}");
            _currentContext.isItalic = xmltag.isStartTag();
            break;
            
        //***** Boldface (fett) *****
        case HtmlTag.B: 
        case HtmlTag.BOLD: 
        case HtmlTag.EMPH: 
        case HtmlTag.STRONG: 
            outputRtf(xmltag.isStartTag() ? "{\\b " : "}");
            _currentContext.isBold = xmltag.isStartTag();
            break;
            
        //***** FixedFont (Typewriter) *****
        case HtmlTag.TT: 
        case HtmlTag.CODE: 
        case HtmlTag.KBD: 
        case HtmlTag.Q: 
        case HtmlTag.SAMP: 
        case HtmlTag.XMP: 
            outputRtf(xmltag.isStartTag() ? "{\\f1 " : "}");
            _currentContext.isFixedFont = xmltag.isStartTag();
            break;
            
     		//***** Underlined (Unterstrichen) *****
        case HtmlTag.UNDERLINE: 
        case HtmlTag.U: 
            outputRtf(xmltag.isStartTag() ? "{\\ul " : "}");
            break;

        //***** Subscript (Tiefgestellt) *****
        case HtmlTag.SUB: 
            outputRtf(xmltag.isStartTag() ? "{\\sub " : "}");
            break;
            
        //***** Superscript (Hochgestellt) *****
        case HtmlTag.SUP: 
            outputRtf(xmltag.isStartTag() ? "{\\super " : "}");
            break;
            
        //***** Big Capitals (groe Kapitlchen) *****
        case HtmlTag.BIG: 
            outputRtf(xmltag.isStartTag() ? "{\\caps " : "}");
            _smallcapsActive = false;
            break;
            
        //***** Small Capitals (kleine Kapitlchen) *****
        case HtmlTag.SMALL: 
            outputRtf(xmltag.isStartTag() ? "{\\caps " : "}");
            _smallcapsActive = xmltag.isStartTag();
            break;
            
  		//---------------------------
  		//-----      Absatz     -----
  		//---------------------------
  		case HtmlTag.P:
          closeLine();
          RtfParagraph rtfparagraph = new RtfParagraph(xmltag.getTagAttr("class"));
          if (xmltag.isStartTag()) {
            if (!_tableActive)
              outputRtf(RtfWriter.PARAGRAPH_BEFORE);
            rtfparagraph.processPre(this);
          } 
          else {
            rtfparagraph.processPost(this);
          }
  		    break;

  		//-------------------------------
      //-----      Headings       -----
      //-------------------------------
      case HtmlTag.H1: 
      case HtmlTag.H2: 
      case HtmlTag.H3: 
      case HtmlTag.H4: 
      case HtmlTag.H5: 
      case HtmlTag.H6: 
        if (_tableActive) {
          closeLine();
          if (xmltag.isStartTag()) {
            if (!_tableActive)
              outputRtf(RtfWriter.PARAGRAPH_BEFORE);
            outputRtf(RtfWriter.PARAGRAPH);
          } 
          else {
            outputRtf(RtfWriter.PARAGRAPH_END);
          }
        }
        else {
          int level = 0;
          switch (taginfo.Ident())
          {
            case HtmlTag.H1: 
              level = 1;
              break;
            case HtmlTag.H2: 
              level = 2;
              break;
            case HtmlTag.H3: 
              level = 3;
              break;
            case HtmlTag.H4: 
              level = 4;
              break;
            case HtmlTag.H5: 
              level = 5;
              break;
            case HtmlTag.H6: 
              level = 6;
              break;
          }
          RtfHeading rtfheading = new RtfHeading(level);
         	if (xmltag.isStartTag()) {
            closeLine();
            if (!_tableActive)
              outputRtf(RtfWriter.PARAGRAPH_BEFORE);
            rtfheading.processPre(this);
          }
          else
            rtfheading.processPost(this);
        }
        break;
      
  		//-------------------------------
      //-----      Images         -----
      //-------------------------------
      case HtmlTag.IMG: 
       	if (xmltag.isStartTag()) {
          try {
            RtfImg rtfimg = new RtfImg();
            rtfimg.setSrc(xmltag.getTagAttr("src"));
            rtfimg.setHeight(xmltag.getTagAttr("height"));
            rtfimg.setWidth(xmltag.getTagAttr("width"));
            closeLine();
            if (!_tableActive)
              outputRtf(RtfWriter.PARAGRAPH_BEFORE);
            outputRtf(rtfimg.process(this));
            lineFeed();
          }
          catch (Exception ex) {
            System.out.println("[]: " + ex.toString());
            ex.printStackTrace();
          }
        }
        break;
      
    		//-------------------------------
        //-----      Hyperlinks     -----
        //-------------------------------
        case HtmlTag.A: 
       		if (xmltag.isStartTag()) {
     				String a_href = xmltag.getTagAttr("href");
     				if (a_href != null)
      			{
            	outputRtf("\n{\\field");
              outputRtf("\n{\\*\\fldinst ");
      				outputRtf("\n{ HYPERLINK " + a_href + " }}");
        			outputRtf("\n{\\fldrslt {\\cs15\\ul" + _output.getColor("blue") + " ");
            	_hyperlinkActive = true;
     				}
       		} else	
       		{
            if (_hyperlinkActive)        					
  					outputRtf("}}}");
     				_hyperlinkActive = false;
      		}
       		break;
        		
      	//---------------------------------------------
        //-----      Alignment (Ausrichtung)      -----
      	//---------------------------------------------
   			
        //****** CENTER ******
        case HtmlTag.CENTER: 
            closeLine();
            outputRtf(xmltag.isStartTag() ? "\\qc " : "\\ql ");
            _currentContext.alignment = xmltag.isStartTag() ? Context.ALIGN_CENTER : Context.ALIGN_LEFT;
            break;
            
        //****** DIV align=??? ******
        case HtmlTag.DIV: 
            processDiv(xmltag);
            break;
            
    		//---------------------------------------
      	//-----      Breaks (Umbrche)      -----
    		//---------------------------------------
   			
        //****** BR ******
        case HtmlTag.BR: 
            closeLine();
            break;
            
        //****** HR ******
        case HtmlTag.HR: 
            closeLine();
            printText("{\\fs10 \\par}");
            printText("{\\fs2 \\ri100 \\brdrb \\brdrs \\par}");
            printText("{\\fs10 \\par}");
            resetToTopOfPage();
            previousLineIsEmpty = false;
            break;

        case HtmlTag.CAPTION: 
            closeLine();
            if(xmltag.isStartTag())
            {
                tableWasActive = _tableActive;
                _tableActive = false;
                outputRtf("\\qc ");
            } else
            {
                lineFeed();
                outputRtf("\\ql ");
                _tableActive = tableWasActive;
            }
            break;

        case HtmlTag.TITLE: 
        case HtmlTag.COMMENT: 
        case HtmlTag.EXCLAM_CMT: 
            _currentContext.discard_text = xmltag.isStartTag();
            break;

        //----------------------------
        //-----      Listen      -----
        //----------------------------
        	
        //*****  UL (Unordered List)  *****
        case HtmlTag.UL: 
        case HtmlTag.OL: 
        case HtmlTag.MENU: 
            if (xmltag.isStartTag()) {
              closeLine();
              if (!_tableActive)
                outputRtf(RtfWriter.PARAGRAPH_BEFORE);
              _currentContext.list_offset += ConversionTemplate.LST_OFFSET;
              _currentContext.list_type = taginfo.Ident() != HtmlTag.OL ? Context.LIST_UL : Context.LIST_OL;
              _currentContext.list_number = _output.getListNumber(_currentContext.list_type);
            } 
            else {
              closeLine();
              _currentContext.list_offset -= ConversionTemplate.LST_OFFSET;
              if(_currentContext.list_offset < 0)
                  _currentContext.list_offset = 0;
              outputRtf(RtfWriter.PARD + "\\li" + _currentContext.list_offset + " ");
              lineFeed();
            }
            break;

        //*****  LI (List Item)  *****
        case HtmlTag.LI: 
        case HtmlTag.DD: 
            if(Context != null && Context.current_tag == HtmlTag.DT)
                closeOpenTag("DT");
            if(xmltag.isStartTag())
            {
                if(Context != null && Context.current_tag == HtmlTag.DD)
                    closeOpenTag("DD");
                if(Context != null && Context.current_tag == HtmlTag.LI)
                    closeOpenTag("LI");
                closeLine();
                if(_currentContext.list_offset >= 0)
                    outputRtf(RtfWriter.PARD + "\\li" + _currentContext.list_offset + " ");
                switch(_currentContext.list_type)
                {
                case RtfGeneratorContext.LIST_OL: 
                case RtfGeneratorContext.LIST_UL: 
   		        	_output.printListItemStart(_currentContext.list_type,_currentContext.list_number);
					break;
           		default:
                  	if(taginfo.Ident() == HtmlTag.LI)
                    	_output.printListItemStart(Context.LIST_UL,_currentContext.list_number);
  		            else
      		            nextWordOnBol = true;
					break;
                }
            } else
            {
           		printText(_output.ENDGROUP);
                if(_currentContext.list_type == Context.LIST_NO)
                    outputRtf(RtfWriter.PARD + "\\li0 ");
            }
            break;

        case HtmlTag.DL: 
            if(Context != null && Context.current_tag == HtmlTag.DD)
                closeOpenTag("DD");
            if(Context != null && Context.current_tag == HtmlTag.DT)
                closeOpenTag("DT");
            if(Context != null && Context.current_tag == HtmlTag.LI)
                closeOpenTag("LI");
            if(xmltag.isStartTag())
            {
                lineFeed();
                _currentContext.list_offset += ConversionTemplate.LST_OFFSET;
                _currentContext.list_type = Context.LIST_DL;
                nextWordOnBol = true;
            } else
            {
                lineFeed();
                _currentContext.list_offset -= ConversionTemplate.LST_OFFSET;
                if(_currentContext.list_offset < 0)
                    _currentContext.list_offset = 0;
                outputRtf(RtfWriter.PARD + "\\li" + _currentContext.list_offset + " ");
            }
            break;

        case HtmlTag.DT: 
            if(Context != null && Context.current_tag == HtmlTag.DD)
                closeOpenTag("DD");
            if(xmltag.isStartTag())
            {
                if(Context != null && Context.current_tag == HtmlTag.DT)
                    closeOpenTag("DT");
                closeLine();
                int k = _currentContext.list_offset - ConversionTemplate.LST_OFFSET;
                if(k >= 0)
                    outputRtf(RtfWriter.PARD + "\\li" + k + " ");
            } else
            {
                closeLine();
                if(_currentContext.list_type == Context.LIST_NO)
                    outputRtf(RtfWriter.PARD + "\\li0 ");
            }
            break;

     		//-----------------------------
      	//-----      Tabelle      -----
        //-----------------------------
   			
        //*****  TABLE  *****
        case HtmlTag.TABLE: 
            if(xmltag.isStartTag())
            {
                if (_currentTable != null) {
                  tableStack.push(_currentTable);
                }
                else {
                  closeLine();
                  if (!_tableActive)
                    outputRtf(RtfWriter.PARAGRAPH_BEFORE);
                }
                int align = _currentContext.alignment != Context.ALIGN_CENTER 
                					? _currentContext.alignment != Context.ALIGN_RIGHT 
                					? Context.ALIGN_LEFT 
                					: Context.ALIGN_RIGHT 
                					: Context.ALIGN_CENTER;
                _currentTable = new RtfTable(xmltag, _currentContext.list_offset, align, _output);
            } else
            if(_currentTable != null)
            {
                if(!tableStack.empty())
                {
                    RtfTable rtftable = (RtfTable)tableStack.pop();
                    rtftable.addTableToCell(_currentTable);
                    _currentTable = rtftable;
                } else
                {
                    _currentTable.dumpRTF();
                    _currentTable = null;
                    _tableActive = false;
                    tableWasActive = false;
                    outputRtf(RtfWriter.PARD + "\\li" + _currentContext.list_offset + " ");
                }
            }
            break;

        //*****  TR (Table Row)  *****
        case HtmlTag.TR: 
            if (xmltag.isStartTag() && _currentTable != null) {
              _tableActive = true;
              _currentTable.addRow(xmltag);
            }
            break;

        //*****  TH (Table Head)  *****
        //*****  TD (Table Data)  *****
        case HtmlTag.TH: 
        case HtmlTag.TD: 
            boolean isStartTag = previousLineIsEmpty;
            if (xmltag.isStartTag() && _currentTable != null)
            {
                if(!_tableActive)
                    processTag(new Tag("<TR>"));
                _currentTable.addCell(xmltag, taginfo.Ident() == HtmlTag.TH);
            }
            resetToTopOfPage();
            if(xmltag.isStartTag())
                previousLineIsEmpty = isStartTag;
            else
           if(!xmltag.isStartTag())
                previousLineIsEmpty = false;
            break;
        }

        //-----------------------------
        //-----------------------------
        
        if(!xmltag.isStartTag() || !taginfo.needsCloseTag())
        {
            if(!_contextStack.empty())
              _currentContext = (RtfGeneratorContext)_contextStack.pop();
            else
            	System.out.println("[RtfGenerator.processTag]: Warning -- Tag stack is empty. Some tags were not closed.");
        }
    }

    

  /**
  * closeOpenTag
  * @param  tag
  */    
  private void closeOpenTag(String tag)
  {
    RtfGeneratorContext Context = (RtfGeneratorContext)_contextStack.pop();
  	processTag(new Tag("</" + tag + ">"));
    _contextStack.push(Context);
  }

  /**
  * resetToTopOfPage
  */    
  private void resetToTopOfPage()
  {
    nextWordOnBol = true;
  	currentLineClosed = true;
    previousLineIsEmpty = true;
  }
    
  /**
  * dumpText
  */    
private void dumpText()
{
	if(_text.length() == 0)
		return;
            
	//-----  Discard text? (e.g. HEAD, COMMENT; etc.)  -----
	if (_currentContext.discard_text)
	{
		_text = "";
		return;
	}
        
	//-----  Small capitals? Convert buffer to lowercase  -----
	if (_smallcapsActive)
        _text = _text.toLowerCase();
        	
	//-----  Remove characters  -----
	int i = 0;
	//* remove multiple-blanks from text *
	while((i = _text.indexOf("  ", i)) >= 0) 
		_text = _text.substring(0, i) + _text.substring(i + 1);
            			
	//-----  Print buffer to RTF  -----
	printText(_text);
        
	//-----  Reset buffer  -----
	_text = "";
}

/*	######################################################################
	###  printText
	###
	######################################################################
*/
private void printText(String s)
{
	s = HtmlObject.translateSpecialChars(s);
	outputRtf(s);
	checkCurrentLineStatus(s);
}
    
/*	######################################################################
	###  checkCurrentLineStatus
	###
	######################################################################
*/
private void checkCurrentLineStatus(String s)
{
	s = s.trim();
	if (s.length() > 0)
		if(s.endsWith("\\par") || s.endsWith("\\line"))
		{
			currentLineClosed = true;
			previousLineIsEmpty = nextWordOnBol && (s.equals("\\par") || s.equals("\\line"));
			nextWordOnBol = true;
		} else
		{
			currentLineClosed = false;
			nextWordOnBol = false;
		}
}

/*	######################################################################
	###  processDiv
	###
	######################################################################
*/
private void processDiv(Tag tag)
{
	closeLine();
	String align = tag.getTagAttr("align");
	if (align != null)
	{
		align = align.toUpperCase();
		if (align.equals("RIGHT"))
		{
			_currentContext.alignment = RtfGeneratorContext.ALIGN_RIGHT;
			outputRtf("\\qr ");
			return;
		}
		if(align.equals("CENTER"))
		{
			_currentContext.alignment = RtfGeneratorContext.ALIGN_CENTER;
			outputRtf("\\qc ");
			return;
		}
		if(align.equals("LEFT"))
		{
			_currentContext.alignment = RtfGeneratorContext.ALIGN_LEFT;
			outputRtf("\\ql ");
			return;
		}
	} else
	if(!_tableActive)
	{
		outputRtf("\\ql ");
		_currentContext.alignment = RtfGeneratorContext.ALIGN_LEFT;
	}
}

/*	######################################################################
	###  inStack
	###
	######################################################################
*/
private boolean inStack(String s)
{
	for(Stack stack = (Stack)_contextStack.clone(); !stack.empty();)
	{
		RtfGeneratorContext context = (RtfGeneratorContext)stack.pop();
		if (context.tag_str.equals(s))
			return true;
	}
	return false;
}

  /**
	* lineFeed
  */
  public void lineFeed()
  {
    closeLine();
  	if (!previousLineIsEmpty)
    	printText(RtfWriter.EOL);
  }
	
  /**
	* outputRtf
  */
  public void outputRtf(String s)
  {
    if (_tableActive) {
      System.out.println("[outputRtf]: Add to table = " + s);
      _currentTable.addTextToCell(s);
    }
    else {
      _output.write(s);
    }
  }

/*	######################################################################
	###  closeLine
	###
	######################################################################
*/
public void closeLine()
{
	if(!currentLineClosed)
	{
		printText(RtfWriter.EOL);
		if (_tableActive)
			_currentTable.insertLineBreak();
	}
}

}

